@@ -14,9 +14,9 @@ from .evaluation.eval_stats import eval_stats_classification, eval_stats_regress
 from .normalisation import NormalisationMode
 from .tensor_model import TensorToTensorRegressionModel, TensorToScalarRegressionModel, \
     TensorToTensorClassificationModel, TensorToScalarClassificationModel
-from .vector_model import PredictorModel, VectorModel, VectorRegressionModel, VectorClassificationModel
+from .vector_model import VectorModelBase, VectorModel, VectorRegressionModel, VectorClassificationModel
 
-__version__ = "0.1.6"
+__version__ = "0.1.8"
 
 # The following submodules are not imported by default to avoid necessarily requiring their dependencies:
 # tensorflow
@@ -24,3 +24,4 @@ __version__ = "0.1.6"
 # lightgbm
 # catboost
 # geoanalytics
+# xgboost
