@@ -1,5 +1,5 @@
 import logging
-from typing import Union, Optional, Dict
+from typing import Union, Optional
 
 import sklearn.ensemble
 import sklearn.linear_model
@@ -7,27 +7,55 @@ import sklearn.neighbors
 import sklearn.neural_network
 import sklearn.svm
 
-from .sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnMultiDimVectorRegressionModel
-
+from .sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnMultiDimVectorRegressionModel, \
+    FeatureImportanceProviderSkLearnRegressionMultipleOneDim, FeatureImportanceProviderSkLearnRegressionMultiDim
 
 log = logging.getLogger(__name__)
 
 
-class SkLearnRandomForestVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
+class SkLearnRandomForestVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel, FeatureImportanceProviderSkLearnRegressionMultipleOneDim):
     def __init__(self, n_estimators=100, min_samples_leaf=10, random_state=42, **modelArgs):
         super().__init__(sklearn.ensemble.RandomForestRegressor,
             n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, random_state=random_state, **modelArgs)
 
-    def getFeatureImportances(self) -> Dict[str, Dict[str, float]]:
-        return {targetFeature: dict(zip(self._modelInputVariableNames, model.feature_importances_)) for targetFeature, model in self.models.items()}
+
+class SkLearnLinearRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel, FeatureImportanceProviderSkLearnRegressionMultiDim):
+    def __init__(self, fit_intercept=True, **modelArgs):
+        """
+        :param fit_intercept: whether to determine the intercept, i.e. the constant term which is not scaled with an input feature value;
+            set to False if the data is already centred
+        :param modelArgs: see https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html
+        """
+        super().__init__(sklearn.linear_model.LinearRegression, fit_intercept=fit_intercept, **modelArgs)
 
 
-class SkLearnLinearRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
-    def __init__(self, **modelArgs):
-        super().__init__(sklearn.linear_model.LinearRegression, **modelArgs)
+class SkLearnLinearRidgeRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel, FeatureImportanceProviderSkLearnRegressionMultiDim):
+    """
+    Linear least squares with L2 regularisation
+    """
+    def __init__(self, alpha=1.0, fit_intercept=True, solver="auto", max_iter=None, tol=1e-3, **modelArgs):
+        """
+        :param alpha: multiplies the L2 term, controlling regularisation strength
+        :param fit_intercept: whether to determine the intercept, i.e. the constant term which is not scaled with an input feature value;
+            set to False if the data is already centred
+        :param modelArgs: see https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html#sklearn.linear_model.Ridge
+        """
+        super().__init__(sklearn.linear_model.Ridge, alpha=alpha, fit_intercept=fit_intercept, max_iter=max_iter, tol=tol,
+            solver=solver, **modelArgs)
 
-    def getFeatureImportances(self) -> Dict[str, float]:
-        return dict(zip(self._modelInputVariableNames, self.model.feature_importances_))
+
+class SkLearnLinearLassoRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel, FeatureImportanceProviderSkLearnRegressionMultiDim):
+    """
+    Linear least squares with L1 regularisation, a.k.a. the lasso
+    """
+    def __init__(self, alpha=1.0, fit_intercept=True, max_iter=1000, tol=0.0001, **modelArgs):
+        """
+        :param alpha: multiplies the L1 term, controlling regularisation strength
+        :param fit_intercept: whether to determine the intercept, i.e. the constant term which is not scaled with an input feature value;
+            set to False if the data is already centred
+        :param modelArgs: see https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html#sklearn.linear_model.Lasso
+        """
+        super().__init__(sklearn.linear_model.Lasso, alpha=alpha, fit_intercept=fit_intercept, max_iter=max_iter, tol=tol, **modelArgs)
 
 
 class SkLearnMultiLayerPerceptronVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
@@ -75,3 +103,9 @@ class SkLearnExtraTreesVectorRegressionModel(AbstractSkLearnMultipleOneDimVector
     def __init__(self, n_estimators=100, min_samples_leaf=10, random_state=42, **modelArgs):
         super().__init__(sklearn.ensemble.ExtraTreesRegressor,
             n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, random_state=random_state, **modelArgs)
+
+
+class SkLearnDummyVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
+    def __init__(self, strategy='mean', constant=None, quantile=None):
+        super().__init__(sklearn.dummy.DummyRegressor,
+            strategy=strategy, constant=constant, quantile=quantile)
