@@ -6,19 +6,19 @@ import time
 from abc import ABC, abstractmethod
 from collections import OrderedDict
 from enum import Enum
-from typing import List, Union, Sequence, Callable, TYPE_CHECKING, Tuple, Optional
+from typing import List, Union, Sequence, Callable, TYPE_CHECKING, Tuple, Optional, Dict
 
 import matplotlib.figure
-from matplotlib import pyplot as plt
 import numpy as np
 import pandas as pd
 import torch
 import torch.nn as nn
 import torch.optim as optim
+from matplotlib import pyplot as plt
 from torch import cuda as torchcuda
 
 from .torch_data import TensorScaler, DataUtil, TorchDataSet, TorchDataSetProviderFromDataUtil, TorchDataSetProvider, \
-    TensorScalerIdentity
+    TensorScalerIdentity, TensorTransformer
 from .torch_enums import ClassificationOutputMode
 from ..util.string import ToStringMixin
 
@@ -41,13 +41,20 @@ class Optimiser(enum.Enum):
     LBFGS = ("lbfgs", optim.LBFGS)
 
     @classmethod
-    def fromName(cls, name: str):
+    def fromName(cls, name: str) -> "Optimiser":
         lname = name.lower()
         for o in cls:
             if o.value[0] == lname:
                 return o
         raise ValueError(f"Unknown optimiser name '{name}'; known names: {[o.value[0] for o in cls]}")
 
+    @classmethod
+    def fromNameOrInstance(cls, nameOrInstance: Union[str, "Optimiser"]) -> "Optimiser":
+        if type(nameOrInstance) == str:
+            return cls.fromName(nameOrInstance)
+        else:
+            return nameOrInstance
+
 
 class _Optimiser(object):
     """
@@ -61,10 +68,7 @@ class _Optimiser(object):
         :param max_grad_norm: gradient norm value beyond which to apply gradient shrinkage
         :param optimiserArgs: keyword arguments to be used in actual torch optimiser
         """
-        if type(method) == str:
-            self.method = Optimiser.fromName(method)
-        else:
-            self.method = method
+        self.method = Optimiser.fromNameOrInstance(method)
         self.params = list(params)  # careful: params may be a generator
         self.last_ppl = None
         self.lr = lr
@@ -106,6 +110,143 @@ class NNLossEvaluator(ABC):
     """
     Base class defining the interface for training and validation loss evaluation.
     """
+    class Evaluation(ABC):
+        @abstractmethod
+        def startEpoch(self) -> None:
+            """
+            Starts a new epoch, resetting any aggregated values required to ultimately return the
+            epoch's overall training loss (via getEpochTrainLoss) and validation metrics (via getValidationMetrics)
+            """
+            pass
+
+        @abstractmethod
+        def computeTrainBatchLoss(self, modelOutput, groundTruth, X, Y) -> torch.Tensor:
+            """
+            Computes the loss for the given model outputs and ground truth values for a batch
+            and aggregates the computed loss values such that :meth:``getEpochTrainLoss`` can return an appropriate
+            result for the entire epoch.
+            The original batch tensors X and Y are provided as meta-information only.
+
+            :param modelOutput: the model output
+            :param groundTruth: the ground truth values
+            :param X: the original batch input tensor
+            :param Y: the original batch output (ground truth) tensor
+            :return: the loss (scalar tensor)
+            """
+            pass
+
+        @abstractmethod
+        def getEpochTrainLoss(self) -> float:
+            """
+            :return: the epoch's overall training loss (as obtained by collecting data from individual training
+                batch data passed to computeTrainBatchLoss)
+            """
+            pass
+
+        @abstractmethod
+        def processValidationBatch(self, modelOutput, groundTruth, X, Y) -> None:
+            """
+            Processes the given model outputs and ground truth values in order to compute sufficient statistics for
+            velidation metrics, which at the end of the epoch, shall be retrievable via method getValidationMetrics
+
+            :param modelOutput: the model output
+            :param groundTruth: the ground truth values
+            :param X: the original batch input tensor
+            :param Y: the original batch output (ground truth) tensor
+            :return: the loss (scalar tensor)
+            """
+            pass
+
+        @abstractmethod
+        def getValidationMetrics(self) -> Dict[str, float]:
+            pass
+
+    @abstractmethod
+    def startEvaluation(self, cuda: bool) -> Evaluation:
+        """
+        Begins the evaluation of a model, returning a (stateful) object which is to perform the necessary computations.
+
+        :param cuda: whether CUDA is being applied (all tensors/models on the GPU)
+        :return: the evaluation object
+        """
+        pass
+
+    @abstractmethod
+    def getValidationMetricName(self) -> str:
+        """
+        :return: the name of the validation metric which is to be used to determine the best model (key for the ordered
+            dictionary returned by method Evaluation.getValidationMetrics)
+        """
+        pass
+
+
+class NNLossEvaluatorFixedDim(NNLossEvaluator, ABC):
+    """
+    Base class defining the interface for training and validation loss evaluation, which uses fixed-dimension
+    outputs and aggregates individual training batch losses that are summed losses per batch
+    (averaging appropriately internally).
+    """
+    class Evaluation(NNLossEvaluator.Evaluation):
+        def __init__(self, criterion, validationLossEvaluator: "NNLossEvaluatorFixedDim.ValidationLossEvaluator",
+                outputDimWeights: torch.Tensor = None):
+            self.outputDimWeights = outputDimWeights
+            self.outputDimWeightSum = torch.sum(outputDimWeights) if outputDimWeights is not None else None
+            self.validationLossEvaluator = validationLossEvaluator
+            self.criterion = criterion
+            self.totalLoss = None
+            self.numSamples = None
+            self.numOutputsPerDataPoint: Optional[int] = None
+            self.validationGroundTruthShape = None
+
+        def startEpoch(self):
+            self.totalLoss = 0
+            self.numSamples = 0
+            self.validationGroundTruthShape = None
+
+        def computeTrainBatchLoss(self, modelOutput, groundTruth, X, Y) -> torch.Tensor:
+            # size of modelOutput and groundTruth: (batchSize, outputDim=numOutputsPerDataPoint)
+            if self.numOutputsPerDataPoint is None:
+                outputShape = Y.shape[1:]
+                self.numOutputsPerDataPoint = functools.reduce(lambda x, y: x * y, outputShape, 1)
+                assert self.outputDimWeights is None or len(self.outputDimWeights) == self.numOutputsPerDataPoint
+            numDataPointsInBatch = Y.shape[0]
+            if self.outputDimWeights is None:
+                # treat all dimensions as equal, applying criterion to entire tensors
+                loss = self.criterion(modelOutput, groundTruth)
+                self.numSamples += numDataPointsInBatch * self.numOutputsPerDataPoint
+                self.totalLoss += loss.item()
+                return loss
+            else:
+                # compute loss per dimension and return weighted loss
+                lossPerDim = torch.zeros(self.numOutputsPerDataPoint, device=modelOutput.device, dtype=torch.float)
+                for o in range(self.numOutputsPerDataPoint):
+                    lossPerDim[o] = self.criterion(modelOutput[:, o], groundTruth[:, o])
+                weightedLoss = (lossPerDim * self.outputDimWeights).sum() / self.outputDimWeightSum
+                self.numSamples += numDataPointsInBatch
+                self.totalLoss += weightedLoss.item()
+                return weightedLoss
+
+        def getEpochTrainLoss(self) -> float:
+            return self.totalLoss / self.numSamples
+
+        def processValidationBatch(self, modelOutput, groundTruth, X, Y):
+            if self.validationGroundTruthShape is None:
+                self.validationGroundTruthShape = Y.shape[1:]  # the shape of the output of a single model application
+                self.validationLossEvaluator.startValidationCollection(self.validationGroundTruthShape)
+            self.validationLossEvaluator.processValidationResultBatch(modelOutput, groundTruth)
+
+        def getValidationMetrics(self) -> Dict[str, float]:
+            return self.validationLossEvaluator.endValidationCollection()
+
+    def startEvaluation(self, cuda: bool) -> Evaluation:
+        criterion = self.getTrainingCriterion()
+        outputDimWeightsArray = self.getOutputDimWeights()
+        outputDimWeightsTensor = torch.from_numpy(outputDimWeightsArray).float() if outputDimWeightsArray is not None else None
+        if cuda:
+            criterion = criterion.cuda()
+            if outputDimWeightsTensor is not None:
+                outputDimWeightsTensor = outputDimWeightsTensor.cuda()
+        return self.Evaluation(criterion, self.createValidationLossEvaluator(cuda), outputDimWeights=outputDimWeightsTensor)
 
     @abstractmethod
     def getTrainingCriterion(self) -> nn.Module:
@@ -115,6 +256,10 @@ class NNLossEvaluator(ABC):
         """
         pass
 
+    @abstractmethod
+    def getOutputDimWeights(self) -> Optional[np.ndarray]:
+        pass
+
     @abstractmethod
     def createValidationLossEvaluator(self, cuda: bool) -> "ValidationLossEvaluator":
         """
@@ -133,7 +278,6 @@ class NNLossEvaluator(ABC):
         """
         pass
 
-
     class ValidationLossEvaluator(ABC):
         @abstractmethod
         def startValidationCollection(self, groundTruthShape):
@@ -166,7 +310,7 @@ class NNLossEvaluator(ABC):
             pass
 
 
-class NNLossEvaluatorRegression(NNLossEvaluator):
+class NNLossEvaluatorRegression(NNLossEvaluatorFixedDim, ToStringMixin):
     """A loss evaluator for (multi-variate) regression."""
 
     class LossFunction(Enum):
@@ -175,7 +319,24 @@ class NNLossEvaluatorRegression(NNLossEvaluator):
         MSELOSS = "MSELoss"
         SMOOTHL1LOSS = "SmoothL1Loss"
 
-    def __init__(self, lossFn: LossFunction = LossFunction.L2LOSS):
+    def __init__(self, lossFn: LossFunction = LossFunction.L2LOSS, validationTensorTransformer: Optional[TensorTransformer] = None,
+            outputDimWeights: Sequence[float] = None, applyOutputDimWeightsInValidation=True, validationMetricName: Optional[str] = None):
+        """
+        :param lossFn: the loss function to use
+        :param validationTensorTransformer: a transformer which is to be applied to validation tensors (both model outputs and ground truth)
+            prior to computing the validation metrics
+        :param outputDimWeights: vector of weights to apply to then mean loss per output dimension, i.e. for the case where for each data point,
+            the model produces n output dimensions, the mean loss for the i-th dimension is to be computed separately and be scaled with
+            the weight, and the overall loss returned is the weighted average. The weights need not sum to 1 (normalisation is applied).
+        :param applyOutputDimWeightsInValidation: whether output dimension weights are also to be applied to to the metrics computed
+            for validation. Note that this may not be possible if a validationTensorTransformer which changes the output dimensions is
+            used.
+        :param validationMetricName: the metric to use for model selection during validation; if None, use default depending on lossFn
+        """
+        self.validationTensorTransformer = validationTensorTransformer
+        self.outputDimWeights = np.array(outputDimWeights) if outputDimWeights is not None else None
+        self.applyOutputDimWeightsInValidation = applyOutputDimWeightsInValidation
+        self.validationMetricName = validationMetricName
         if lossFn is None:
             lossFn = self.LossFunction.L2LOSS
         try:
@@ -183,11 +344,9 @@ class NNLossEvaluatorRegression(NNLossEvaluator):
         except ValueError:
             raise Exception(f"The loss function '{lossFn}' is not supported. Available options are: {[e.value for e in self.LossFunction]}")
 
-    def __str__(self):
-        return f"{self.__class__.__name__}[{self.lossFn}]"
-
     def createValidationLossEvaluator(self, cuda):
-        return self.ValidationLossEvaluator(cuda)
+        return self.ValidationLossEvaluator(cuda, self.validationTensorTransformer, self.outputDimWeights,
+            self.applyOutputDimWeightsInValidation)
 
     def getTrainingCriterion(self):
         if self.lossFn is self.LossFunction.L1LOSS:
@@ -200,8 +359,15 @@ class NNLossEvaluatorRegression(NNLossEvaluator):
             raise AssertionError(f"Loss function {self.lossFn} defined but instantiation not implemented.")
         return criterion
 
-    class ValidationLossEvaluator(NNLossEvaluator.ValidationLossEvaluator):
-        def __init__(self, cuda: bool):
+    def getOutputDimWeights(self) -> Optional[np.ndarray]:
+        return self.outputDimWeights
+
+    class ValidationLossEvaluator(NNLossEvaluatorFixedDim.ValidationLossEvaluator):
+        def __init__(self, cuda: bool, validationTensorTransformer: Optional[TensorTransformer], outputDimWeights: np.ndarray,
+                applyOutputDimWeights: bool):
+            self.validationTensorTransformer = validationTensorTransformer
+            self.outputDimWeights = outputDimWeights
+            self.applyOutputDimWeights = applyOutputDimWeights
             self.total_loss_l1 = None
             self.total_loss_l2 = None
             self.outputDims = None
@@ -211,16 +377,29 @@ class NNLossEvaluatorRegression(NNLossEvaluator):
             if cuda:
                 self.evaluateL1 = self.evaluateL1.cuda()
                 self.evaluateL2 = self.evaluateL2.cuda()
+            self.beginNewValidationCollection: Optional[bool] = None
 
         def startValidationCollection(self, groundTruthShape):
             if len(groundTruthShape) != 1:
                 raise ValueError("Outputs that are not vectors are currently unsupported")
-            self.outputDims = groundTruthShape[-1]
-            self.total_loss_l1 = np.zeros(self.outputDims)
-            self.total_loss_l2 = np.zeros(self.outputDims)
-            self.allTrueOutputs = None
+            self.beginNewValidationCollection = True
 
         def processValidationResultBatch(self, output, groundTruth):
+            # apply tensor transformer (if any)
+            if self.validationTensorTransformer is not None:
+                output = self.validationTensorTransformer.transform(output)
+                groundTruth = self.validationTensorTransformer.transform(groundTruth)
+
+            # check if new collection
+            if self.beginNewValidationCollection:
+                self.outputDims = groundTruth.shape[-1]
+                self.total_loss_l1 = np.zeros(self.outputDims)
+                self.total_loss_l2 = np.zeros(self.outputDims)
+                self.allTrueOutputs = None
+                self.beginNewValidationCollection = False
+
+            assert len(output.shape) == 2 and len(groundTruth.shape) == 2
+
             # obtain series of outputs per output dimension: (batch_size, output_size) -> (output_size, batch_size)
             predictedOutput = output.permute(1, 0)
             trueOutput = groundTruth.permute(1, 0)
@@ -254,20 +433,29 @@ class NNLossEvaluatorRegression(NNLossEvaluator):
                 rrse[i] = np.sqrt(mse[i]) / np.sqrt(
                     refModelSumSquaredErrors / numSamples) if refModelSumSquaredErrors != 0 else np.inf
 
-            metrics = OrderedDict([("RRSE", np.mean(rrse)), ("RAE", np.mean(rae)), ("MSE", np.mean(mse)), ("MAE", np.mean(mae))])
+            def mean(x):
+                if self.applyOutputDimWeights:
+                    return np.average(x, weights=self.outputDimWeights)
+                else:
+                    return np.mean(x)
+
+            metrics = OrderedDict([("RRSE", mean(rrse)), ("RAE", mean(rae)), ("MSE", mean(mse)), ("MAE", mean(mae))])
             return metrics
 
     def getValidationMetricName(self):
-        if self.lossFn is self.LossFunction.L1LOSS or self.lossFn is self.LossFunction.SMOOTHL1LOSS:
-            return "MAE"
-        elif self.lossFn is self.LossFunction.L2LOSS or self.lossFn is self.LossFunction.MSELOSS:
-            return "MSE"
+        if self.validationMetricName is not None:
+            return self.validationMetricName
         else:
-            raise AssertionError(f"No selection criterion defined for loss function {self.lossFn}")
+            if self.lossFn is self.LossFunction.L1LOSS or self.lossFn is self.LossFunction.SMOOTHL1LOSS:
+                return "MAE"
+            elif self.lossFn is self.LossFunction.L2LOSS or self.lossFn is self.LossFunction.MSELOSS:
+                return "MSE"
+            else:
+                raise AssertionError(f"No validation metric defined as selection criterion for loss function {self.lossFn}")
 
 
-class NNLossEvaluatorClassification(NNLossEvaluator):
-    """A loss evaluator for (multi-variate) regression"""
+class NNLossEvaluatorClassification(NNLossEvaluatorFixedDim):
+    """A loss evaluator for classification"""
 
     class LossFunction(Enum):
         CROSSENTROPY = "CrossEntropy"
@@ -308,7 +496,10 @@ class NNLossEvaluatorClassification(NNLossEvaluator):
     def getTrainingCriterion(self):
         return self.lossFn.createCriterion()
 
-    class ValidationLossEvaluator(NNLossEvaluator.ValidationLossEvaluator):
+    def getOutputDimWeights(self) -> Optional[np.ndarray]:
+        return None
+
+    class ValidationLossEvaluator(NNLossEvaluatorFixedDim.ValidationLossEvaluator):
         def __init__(self, cuda: bool, lossFn: "NNLossEvaluatorClassification.LossFunction"):
             self.lossFn = lossFn
             self.totalLoss = None
@@ -351,7 +542,7 @@ class NNOptimiserParams(ToStringMixin):
         """
         :param lossEvaluator: the loss evaluator to use
         :param gpu: the index of the GPU to be used (if CUDA is enabled for the model to be trained); if None, default to first GPU
-        :param optimiser: the name of the optimizer to be used; defaults to "adam"
+        :param optimiser: the optimiser to use
         :param optimiserLR: the optimiser's learning rate
         :param earlyStoppingEpochs: the number of epochs without validation score improvement after which to abort training and
             use the best epoch's model (early stopping); if None, never abort training before all epochs are completed
@@ -361,12 +552,12 @@ class NNOptimiserParams(ToStringMixin):
             If no validation is to be performed, pass 1.0.
         :param scaledOutputs: whether to scale all outputs, resulting in computations of the loss function based on scaled values rather than normalised values.
             Enabling scaling may not be appropriate in cases where there are multiple outputs on different scales/with completely different units.
-        :param useShrinkage: whether to apply shrinkage to gradients whose norm exceeds optimiserClip
-        :param shrinkageClip: the maximum gradient norm beyond which to apply shrinkage (if useShrinkage is True)
+        :param useShrinkage: whether to apply shrinkage to gradients whose norm exceeds ``shrinkageClip``, scaling the gradient down to ``shrinkageClip``
+        :param shrinkageClip: the maximum gradient norm beyond which to apply shrinkage (if ``useShrinkage`` is True)
         :param shuffle: whether to shuffle the training data
         :param optimiserArgs: keyword arguments to be passed on to the actual torch optimiser
         """
-        if optimiser == 'lbfgs':
+        if Optimiser.fromNameOrInstance(optimiser) == Optimiser.LBFGS:
             largeBatchSize = 1e12
             if batchSize is not None:
                 log.warning(f"LBFGS does not make use of batches, therefore using large batch size {largeBatchSize} to achieve use of a single batch")
@@ -427,33 +618,19 @@ class NNOptimiser:
 
     def __init__(self, params: NNOptimiserParams):
         """
-        :param cuda: whether to use CUDA
-        :param lossEvaluator: the loss evaluator to use
-        :param gpu: index of the gpu to be used (if CUDA is enabled in the model to be trained)
-        :param optimiser: the optimizer to be used; defaults to "adam"
-        :param optimiserClip: the maximum gradient norm beyond which to apply shrinkage (if useShrinkage is True)
-        :param optimiserLR: the optimiser's learning rate
-        :param batchSize: the batch size to use; for algorithms L-BFGS (optimiser='lbfgs'), which do not use batches, leave this at None.
-            If the algorithm uses batches and None is specified, batch size 64 will be used by default.
-        :param trainFraction: the fraction of the data used for training (with the remainder being used for validation).
-            If no validation is to be performed, pass 1.0.
-        :param scaledOutputs: whether to scale all outputs, resulting in computations of the loss function based on scaled values rather than normalised values.
-            Enabling scaling may not be appropriate in cases where there are multiple outputs on different scales/with completely different units.
-        :param useShrinkage: whether to apply shrinkage to gradients whose norm exceeds optimiserClip
-        :param optimiserArgs: keyword arguments to be passed on to the actual torch optimiser
+        :param params: parameters
         """
         if params.lossEvaluator is None:
             raise ValueError("Must provide a loss evaluator")
 
         self.params = params
-        self.lossEvaluatorState = None
         self.cuda = None
 
     def __str__(self):
         return f"{self.__class__.__name__}[params={self.params}]"
 
     def fit(self, model: "TorchModel", data: Union[DataUtil, List[DataUtil], TorchDataSetProvider, List[TorchDataSetProvider],
-            TorchDataSet, List[TorchDataSet], Tuple[TorchDataSet, TorchDataSet], List[Tuple[TorchDataSet, TorchDataSet]]],
+                                                   TorchDataSet, List[TorchDataSet], Tuple[TorchDataSet, TorchDataSet], List[Tuple[TorchDataSet, TorchDataSet]]],
             createTorchModule=True) -> "TrainingInfo":
         """
         Fits the parameters of the given model to the given data, which can be a list of or single instance of one of the following:
@@ -537,10 +714,6 @@ class NNOptimiser:
         trainingLog(f"Starting training process via {self}")
 
         lossEvaluator = self.params.lossEvaluator
-        criterion = lossEvaluator.getTrainingCriterion()
-
-        if self.cuda:
-            criterion = criterion.cuda()
 
         totalEpochs = None
         best_val = 1e9
@@ -549,7 +722,7 @@ class NNOptimiser:
             max_grad_norm=self.params.shrinkageClip, use_shrinkage=self.params.useShrinkage, **self.params.optimiserArgs)
 
         bestModelBytes = model.getModuleBytes()
-        self.lossEvaluatorState = lossEvaluator.createValidationLossEvaluator(self.cuda)
+        lossEvaluation = lossEvaluator.startEvaluation(self.cuda)
         validationMetricName = lossEvaluator.getValidationMetricName()
         trainingLossValues = []
         validationMetricValues = []
@@ -557,10 +730,11 @@ class NNOptimiser:
             self.log.info(f'Begin training with cuda={self.cuda}')
             self.log.info('Press Ctrl+C to end training early')
             for epoch in range(1, self.params.epochs + 1):
+                lossEvaluation.startEpoch()
                 epoch_start_time = time.time()
 
                 # perform training step, processing all the training data once
-                train_loss = self._train(trainingSets, torchModel, criterion, optim, self.params.batchSize, self.cuda, outputScalers)
+                train_loss = self._train(trainingSets, torchModel, optim, lossEvaluation, self.params.batchSize, outputScalers)
                 trainingLossValues.append(train_loss)
 
                 # perform validation, computing the mean metrics across all validation sets (if more than one),
@@ -570,7 +744,7 @@ class NNOptimiser:
                     metricsSum = None
                     metricsKeys = None
                     for i, (validationSet, outputScaler) in enumerate(zip(validationSets, outputScalers)):
-                        metrics = self._evaluate(validationSet, torchModel, outputScaler)
+                        metrics = self._evaluate(validationSet, torchModel, lossEvaluation, outputScaler)
                         metricsArray = np.array(list(metrics.values()))
                         if i == 0:
                             metricsSum = metricsArray
@@ -616,7 +790,7 @@ class NNOptimiser:
             model.setModuleBytes(bestModelBytes)
 
         return TrainingInfo(bestEpoch=best_epoch if useValidation else None, log=trainingLogEntries, totalEpochs=totalEpochs,
-                trainingLossSequence=trainingLossValues, validationMetricSequence=validationMetricValues)
+            trainingLossSequence=trainingLossValues, validationMetricSequence=validationMetricValues)
 
     def _applyModel(self, model, input: Union[torch.Tensor, Sequence[torch.Tensor]], groundTruth, outputScaler: TensorScaler):
         if isinstance(input, torch.Tensor):
@@ -633,46 +807,31 @@ class NNOptimiser:
         scaledTruth = outputScaler.denormalise(groundTruth)
         return scaledOutput, scaledTruth
 
-    def _train(self, dataSets: Sequence[TorchDataSet], model: nn.Module, criterion: nn.modules.loss._Loss,
-            optim: _Optimiser, batch_size: int, cuda: bool, outputScalers: Sequence[TensorScaler]):
+    def _train(self, dataSets: Sequence[TorchDataSet], model: nn.Module, optim: _Optimiser,
+            lossEvaluation: NNLossEvaluator.Evaluation, batch_size: int, outputScalers: Sequence[TensorScaler]):
         """Performs one training epoch"""
         model.train()
-        total_loss = 0
-        n_samples = 0
-        numOutputsPerDataPoint = None
         for dataSet, outputScaler in zip(dataSets, outputScalers):
             for X, Y in dataSet.iterBatches(batch_size, shuffle=self.params.shuffle):
-                if numOutputsPerDataPoint is None:
-                    outputShape = Y.shape[1:]
-                    numOutputsPerDataPoint = functools.reduce(lambda x, y: x * y, outputShape, 1)
-
                 def closure():
                     model.zero_grad()
                     output, groundTruth = self._applyModel(model, X, Y, outputScaler)
-                    loss = criterion(output, groundTruth)
+                    loss = lossEvaluation.computeTrainBatchLoss(output, groundTruth, X, Y)
                     loss.backward()
                     return loss
 
-                loss = optim.step(closure)
-                total_loss += loss.item()
-                numDataPointsInBatch = Y.size(0)
-                n_samples += numDataPointsInBatch * numOutputsPerDataPoint
-        return total_loss / n_samples
+                optim.step(closure)
+        return lossEvaluation.getEpochTrainLoss()
 
-    def _evaluate(self, dataSet: TorchDataSet, model: nn.Module, outputScaler: TensorScaler):
+    def _evaluate(self, dataSet: TorchDataSet, model: nn.Module, lossEvaluation: NNLossEvaluator.Evaluation,
+            outputScaler: TensorScaler):
         """Evaluates the model on the given data set (a validation set)"""
         model.eval()
-
-        groundTruthShape = None
         for X, Y in dataSet.iterBatches(self.params.batchSize, shuffle=False):
-            if groundTruthShape is None:
-                groundTruthShape = Y.shape[1:]  # the shape of the output of a single model application
-                self.lossEvaluatorState.startValidationCollection(groundTruthShape)
             with torch.no_grad():
                 output, groundTruth = self._applyModel(model, X, Y, outputScaler)
-            self.lossEvaluatorState.processValidationResultBatch(output, groundTruth)
-
-        return self.lossEvaluatorState.endValidationCollection()
+            lossEvaluation.processValidationBatch(output, groundTruth, X, Y)
+        return lossEvaluation.getValidationMetrics()
 
     def _init_cuda(self):
         """Initialises CUDA (for learning) by setting the appropriate device if necessary"""
@@ -692,8 +851,8 @@ class NNOptimiser:
 
 
 class TrainingInfo:
-    def __init__(self, bestEpoch: int = None, log: List[str] = None, trainingLossSequence: Sequence[float] = None, validationMetricSequence:
-            Sequence[float] = None, totalEpochs=None):
+    def __init__(self, bestEpoch: int = None, log: List[str] = None, trainingLossSequence: Sequence[float] = None,
+            validationMetricSequence: Sequence[float] = None, totalEpochs=None):
         self.validationMetricSequence = validationMetricSequence
         self.trainingLossSequence = trainingLossSequence
         self.log = log
@@ -706,15 +865,35 @@ class TrainingInfo:
         self.__dict__ = state
 
     def getTrainingLossSeries(self) -> pd.Series:
-        return pd.Series(self.trainingLossSequence, name="training loss")
+        return self._createSeriesWithOneBasedIndex(self.trainingLossSequence, name="training loss")
 
     def getValidationMetricSeries(self) -> pd.Series:
-        return pd.Series(self.validationMetricSequence, name="validation metric")
+        return self._createSeriesWithOneBasedIndex(self.validationMetricSequence, name="validation metric")
+
+    def _createSeriesWithOneBasedIndex(self, sequence: Sequence, name: str):
+        series = pd.Series(sequence, name=name)
+        series.index += 1
+        return series
 
     def plotAll(self) -> matplotlib.figure.Figure:
         """
         Plots both the sequence of training loss values and the sequence of validation metric values
         """
-        fig = plt.figure()
-        pd.concat([self.getTrainingLossSeries(), self.getValidationMetricSeries()], axis=1).plot()
+        ts = self.getTrainingLossSeries()
+        vs = self.getValidationMetricSeries()
+
+        fig, primaryAx = plt.subplots(1, 1)
+        secondaryAx = primaryAx.twinx()
+
+        trainingLine = primaryAx.plot(ts, color='blue')
+        validationLine = secondaryAx.plot(vs, color='orange')
+        bestEpocLine = primaryAx.axvline(self.bestEpoch, color='black', linestyle='dashed')
+
+        primaryAx.set_xlabel("epoch")
+        primaryAx.set_ylabel(ts.name)
+        secondaryAx.set_ylabel(vs.name)
+
+        primaryAx.legend(trainingLine + validationLine + [bestEpocLine], [ts.name, vs.name, "best epoch"])
+        plt.tight_layout()
+
         return fig
