@@ -11,10 +11,8 @@ import time
 from abc import abstractmethod, ABC
 from typing import Any, Callable, Iterator, List, Optional, TypeVar
 
-import joblib
-
 from .hash import pickleHash
-from .pickle import PickleFailureDebugger
+from .pickle import loadPickle, dumpPickle
 
 log = logging.getLogger(__name__)
 
@@ -64,33 +62,6 @@ class PersistentList(ABC):
         pass
 
 
-def loadPickle(path, backend="pickle"):
-    with open(path, "rb") as f:
-        if backend == "pickle":
-            return pickle.load(f)
-        elif backend == "joblib":
-            return joblib.load(f)
-        else:
-            raise ValueError(f"Unknown backend '{backend}'")
-
-
-def dumpPickle(obj, picklePath, backend="pickle"):
-    dirName = os.path.dirname(picklePath)
-    if dirName != "":
-        os.makedirs(dirName, exist_ok=True)
-    with open(picklePath, "wb") as f:
-        if backend == "pickle":
-            try:
-                pickle.dump(obj, f)
-            except AttributeError as e:
-                failingPaths = PickleFailureDebugger.debugFailure(obj)
-                raise AttributeError(f"Cannot pickle paths {failingPaths} of {obj}: {str(e)}")
-        elif backend == "joblib":
-            joblib.dump(obj, f)
-        else:
-            raise ValueError(f"Unknown backend '{backend}'")
-
-
 class DelayedUpdateHook:
     """
     Ensures that a given function is executed after an update happens, but delay the execution until
@@ -594,17 +565,20 @@ class CachedValueProviderMixin(ABC):
 
 
 def cached(fn: Callable[[], T], picklePath, functionName=None, validityCheckFn: Optional[Callable[[T], bool]] = None,
-        backend="pickle") -> T:
+        backend="pickle", protocol=pickle.HIGHEST_PROTOCOL, load=True, version=None) -> T:
     """
     :param fn: the function whose result is to be cached
     :param picklePath: the path in which to store the cached result
     :param functionName: the name of the function fn (for the case where its __name__ attribute is not
         informative)
     :param validityCheckFn: an optional function to call in order to check whether a cached result is still valid;
-        the function shall return True if the res is still valid and false otherwise. If a cached result is invalid,
+        the function shall return True if the result is still valid and false otherwise. If a cached result is invalid,
         the function fn is called to compute the result and the cached result is updated.
     :param backend: pickle or joblib
-    :return: the res (either obtained from the cache or the function)
+    :param protocol: the pickle protocol version
+    :param load: whether to load a previously persisted result; if False, do not load an old result but store the newly computed result
+    :param version: if not None, previously persisted data will only be returned if it was stored with the same version
+    :return: the result (either obtained from the cache or the function)
     """
     if functionName is None:
         functionName = fn.__name__
@@ -612,17 +586,34 @@ def cached(fn: Callable[[], T], picklePath, functionName=None, validityCheckFn:
     def callFnAndCacheResult():
         res = fn()
         log.info(f"Saving cached result in {picklePath}")
-        dumpPickle(res, picklePath, backend=backend)
+        if version is not None:
+            persistedRes = {"__cacheVersion": version, "obj": res}
+        else:
+            persistedRes = res
+        dumpPickle(persistedRes, picklePath, backend=backend, protocol=protocol)
         return res
 
     if os.path.exists(picklePath):
-        log.info(f"Loading cached result of function '{functionName}' from {picklePath}")
-        result = loadPickle(picklePath, backend=backend)
-        if validityCheckFn is not None:
-            if not validityCheckFn(result):
-                log.info(f"Cached result is no longer valid, recomputing ...")
-                result = callFnAndCacheResult()
-        return result
+        if load:
+            log.info(f"Loading cached result of function '{functionName}' from {picklePath}")
+            result = loadPickle(picklePath, backend=backend)
+            if validityCheckFn is not None:
+                if not validityCheckFn(result):
+                    log.info(f"Cached result is no longer valid, recomputing ...")
+                    result = callFnAndCacheResult()
+            if version is not None:
+                cachedVersion = None
+                if type(result) == dict:
+                    cachedVersion = result.get("__cacheVersion")
+                if cachedVersion != version:
+                    log.info(f"Cached result has incorrect version ({cachedVersion}, expected {version}), recomputing ...")
+                    result = callFnAndCacheResult()
+                else:
+                    result = result["obj"]
+            return result
+        else:
+            log.info(f"Ignoring previously stored result in {picklePath}, calling function '{functionName}' ...")
+            return callFnAndCacheResult()
     else:
         log.info(f"No cached result found in {picklePath}, calling function '{functionName}' ...")
         return callFnAndCacheResult()
@@ -632,18 +623,27 @@ class PickleCached(object):
     """
     Function decorator for caching function results via pickle
     """
-    def __init__(self, cacheBasePath: str, filenamePrefix: str = None, filename: str = None, backend="pickle"):
+    def __init__(self, cacheBasePath: str, filenamePrefix: str = None, filename: str = None, backend="pickle",
+            protocol=pickle.HIGHEST_PROTOCOL, load=True, version=None):
         """
         :param cacheBasePath: the directory where the pickle cache file will be stored
         :param filenamePrefix: a prefix of the name of the cache file to be created, to which the function name and, where applicable,
             a hash code of the function arguments will be appended and ".cache.pickle" will be appended; if None, use "" (if filename
             has not been provided)
-        :param filename: the full file name of the cache file to be created; this is admissible only if the function has no arguments
+        :param filename: the full file name of the cache file to be created; if the function takes arguments, the filename must
+            contain a placeholder '%s' for the argument hash
+        :param backend: the serialisation backend to use (see dumpPickle)
+        :param protocol: the pickle protocol version to use
+        :param load: whether to load a previously persisted result; if False, do not load an old result but store the newly computed result
+        :param version: if not None, previously persisted data will only be returned if it was stored with the same version
         """
         self.filename = filename
         self.cacheBasePath = cacheBasePath
         self.filenamePrefix = filenamePrefix
         self.backend = backend
+        self.protocol = protocol
+        self.load = load
+        self.version = version
 
         if self.filenamePrefix is None:
             self.filenamePrefix = ""
@@ -653,18 +653,27 @@ class PickleCached(object):
     def __call__(self, fn: Callable, *_args, **_kwargs):
 
         def wrapped(*args, **kwargs):
+            hashCodeStr = None
             haveArgs = len(args) > 0 or len(kwargs) > 0
+            if haveArgs:
+                hashCodeStr = pickleHash((args, kwargs))
             if self.filename is None:
-                filename = self.filenamePrefix + fn.__qualname__
-                if haveArgs:
-                    filename += "-" + pickleHash((args, kwargs))
+                filename = self.filenamePrefix + fn.__qualname__.replace(".<locals>.", ".")
+                if hashCodeStr is not None:
+                    filename += "-" + hashCodeStr
                 filename += ".cache.pickle"
             else:
-                if haveArgs:
-                    raise Exception("Function called with arguments but full cache filename specified: specify a cache filename prefix only to account for argument values")
-                filename = self.filename
+                if hashCodeStr is not None:
+                    if not "%s" in self.filename:
+                        raise Exception("Function called with arguments but full cache filename contains no placeholder (%s) for argument hash")
+                    filename = self.filename % hashCodeStr
+                else:
+                    if "%s" in self.filename:
+                        raise Exception("Function without arguments but full cache filename with placeholder (%s) was specified")
+                    filename = self.filename
             picklePath = os.path.join(self.cacheBasePath, filename)
-            return cached(lambda: fn(*args, **kwargs), picklePath, functionName=fn.__name__, backend=self.backend)
+            return cached(lambda: fn(*args, **kwargs), picklePath, functionName=fn.__name__, backend=self.backend, load=self.load,
+                version=self.version)
 
         return wrapped
 
