@@ -1,3 +1,4 @@
+import math
 import scipy.stats
 
 from .string import objectRepr
@@ -25,5 +26,9 @@ class NormalDistribution:
         return objectRepr(self, ["mean", "std", "unitMax"])
 
 
+def sigmoid(x: float):
+    return math.exp(x) / (1 + math.exp(x))
+
+
 def reduceFloatPrecisionDecimals(f: float, decimals: int) -> float:
     return float(format(f, '.%df' % decimals))
\ No newline at end of file
